/*
  Copyright ICARE
  Contributeur : SIX Bruno, Juin 2006
  Bruno.Six@univ-lille1.fr
  
  Ce logiciel est rgi par la licence CeCILL soumise au droit franais et
  respectant les principes de diffusion des logiciels libres. Vous pouvez
  utiliser, modifier et/ou redistribuer ce programme sous les conditions de la
  licence CeCILL telle que diffuse par le CEA, le CNRS et l'INRIA sur le site
  "http://www.cecill.info".

  En contrepartie de l'accessibilit au code source et des droits de copie, de
  modification et de redistribution accords par cette licence, il n'est offert
  aux utilisateurs qu'une garantie limite.  Pour les mmes raisons, seule une
  responsabilit restreinte pse sur l'auteur du programme, le titulaire des
  droits patrimoniaux et les concdants successifs.

  A cet gard, l'attention de l'utilisateur est attire sur les risques associs
  au chargement,  l'utilisation,  la modification et/ou au dveloppement et 
  la reproduction du logiciel par l'utilisateur tant donn sa spcificit de
  logiciel libre, qui peut le rendre complexe  manipuler et qui le rserve donc
   des dveloppeurs et des professionnels avertis possdant des connaissances
  informatiques approfondies.  Les utilisateurs sont donc invits  charger et
  tester l'adquation du logiciel  leurs besoins dans des conditions permettant
  d'assurer la scurit de leurs systmes et ou de leurs donnes et, plus
  gnralement,  l'utiliser et l'exploiter dans les mmes conditions de
  scurit.

  Le fait que vous puissiez accder  cet en-tte signifie que vous avez pris
  connaissance de la licence CeCILL, et que vous en avez accept les termes.
*/

#include <hdfsds.h>

/**
  * \file hdf_get_file_sds.c
  * \brief Extraction de SDS d'un fichier HDF
  *
  * Cette fonction extrait tout ou partie des SDS d'un fichier HDF,
  * avec ou sans leurs attributs, et les retourne dans une structure.
  * Voir hdfsds.h pour la description de cette structure.
  *
  * Elle retourne 0 en cas de succs, 1 en cas d'erreur HDF et 2 pour
  * toute autre erreur
  * ----------------------------------------------------------------------------
  *
  * Spcification: int hdf_get_file_sds(char *file, char *names[], int nb_sds,
  *                                     int att_sw, struct hdfsds *psds[]);
  *
  * ----------------------------------------------------------------------------
  *
  * \author SIX Bruno, CGTD ICARE
  * \author CRI USTL, Domaine Universitaire Scientifique
  * \version 1.0.0
  * \date 2006-06-09
  *
  * ----------------------------------------------------------------------------
  *
  * <b>Modifications, rvisions:</b>
  *
  * ----------------------------------------------------------------------------
  *
  */
int hdf_get_file_sds(
  /** Chemin du fichier HDF  traiter                */ char *file,
  /** Liste des noms des SDS  extraire              */ char *names[],
  /** Nombre de SDS du fichier HDF                   */ int nb_sds,
  /** 1 pour extraire  les attributs du SDS, 0 sinon */ int att_sw,
  /** Structure contenant les SDS extraits           */ struct hdfsds *psds[] )
{

  /* Dclaration des variables standard */
  int
    isds,      /* Compteur sur les SDS  extraire */
    idim;      /* Compteur sur les dimensions du SDS courant*/
  long 
    data_size; /* Taille en octets des donnes d'un SDS */

  /* Dclaration des variables spcifiques HDF */
  int32
    sd_id,     /* Identificateur HDF du fichier */
    sds_id,    /* Identificateur HDF du SDS courant */
    sds_index, /* Indice du SDS courant dans la liste des SDS */
    nb_att,    /* Nombre d'attributs d'un SDS extrait */
    *dims,     /* Valeurs des dimensions d'un SDS extrait */
    *start;    /* Table d'indices de dbut par dim. pour l'extraction d'un SDS */

  /* Allocation de la table contenant les donnes et mtadonnes des SDS demands */
  *psds = malloc(nb_sds*sizeof(struct hdfsds *));

  /* Ouverture du fichier HDF */
  if ((sd_id = SDstart(file, DFACC_READ)) < 0) goto ERR_HDF;

  /* DEBUT DE BOUCLE SUR LES SDS A EXTRAIRE */
  for (isds = 0; isds < nb_sds; isds++) {

    /* Recherche du SDS par son nom */
    if ((sds_index = SDnametoindex(sd_id, names[isds])) < 0) goto ERR_HDF;

    /* Initialisation de l'accs au SDS */
    if ((sds_id = SDselect(sd_id, sds_index)) < 0) goto ERR_HDF;

    /* Rcupration d'informations sur le SDS */
    if (SDgetinfo(sds_id, (*psds)[isds].name, &((*psds)[isds].rnk), dims = (int32 *)malloc(MAX_VAR_DIMS),
		  &((*psds)[isds].typ), &nb_att) < 0) goto ERR_HDF;
    (*psds)[isds].dim = (int32 *)realloc(dims, (*psds)[isds].rnk); 

    /* Rcupration ventuelle des attributs du SDS */
    if (att_sw) { if (get_hdf_obj_att(sds_id, (*psds)[isds].natt = nb_att, &((*psds)[isds].att)) < 0) goto ERR_FCT; }
    else { (*psds)[isds].att = NULL; (*psds)[isds].natt = 0; }

    /* Allocation dynamique des variables ncessaires:
       - allocation d'un tableau start d'entiers tous nuls de taille gale au rang du SDS
       - allocation de la zone sds[isds].data aprs calcul de sa taille
         (la fonction DFKNTsize retourne la taille d'un typ spcifique HDF)
    */
    if ((start = (int32 *)malloc((*psds)[isds].rnk*sizeof(int32))) == NULL) goto ERR_STD;
    memset(start, 0, (*psds)[isds].rnk*sizeof(int32));
    data_size = DFKNTsize((*psds)[isds].typ);
    for (idim = 0; idim < (*psds)[isds].rnk; idim++) data_size = data_size*(*psds)[isds].dim[idim];
    if (((*psds)[isds].data = (VOIDP)malloc(data_size)) == NULL) goto ERR_STD;

    /* Extraction des donnes du SDS */
    if (SDreaddata(sds_id, start, NULL, (*psds)[isds].dim, (*psds)[isds].data) < 0) goto ERR_HDF;

    /* Libration des zones devenues inutiles */
    free(start);

    /* Fin d'accs au SDS */
    if (SDendaccess(sds_id) < 0) goto ERR_HDF;

  } /* FIN DE BOUCLE SUR LES SDS A EXTRAIRE */

  /* Fermeture du fichier HDF */
  if (SDend(sd_id) < 0) goto ERR_HDF;

  /* Fin normale de la fonction */
  return 0;

 ERR_HDF:   /* Fin de la fonction sur une erreur HDF locale */
  HEprint(stderr, 0);
  return 1;

 ERR_STD:   /* Fin de la fonction sur une autre erreur locale */
  perror("ERROR");
  return 2;

 ERR_FCT:   /* Fin de la fonction sur une erreur dans une fonction appele */
  return 3;
}
