/*
  Copyright ICARE
  Contributeur : SIX Bruno, Juin 2006
  Bruno.Six@univ-lille1.fr
  
  Ce logiciel est rgi par la licence CeCILL soumise au droit franais et
  respectant les principes de diffusion des logiciels libres. Vous pouvez
  utiliser, modifier et/ou redistribuer ce programme sous les conditions de la
  licence CeCILL telle que diffuse par le CEA, le CNRS et l'INRIA sur le site
  "http://www.cecill.info".

  En contrepartie de l'accessibilit au code source et des droits de copie, de
  modification et de redistribution accords par cette licence, il n'est offert
  aux utilisateurs qu'une garantie limite.  Pour les mmes raisons, seule une
  responsabilit restreinte pse sur l'auteur du programme, le titulaire des
  droits patrimoniaux et les concdants successifs.

  A cet gard, l'attention de l'utilisateur est attire sur les risques associs
  au chargement,  l'utilisation,  la modification et/ou au dveloppement et 
  la reproduction du logiciel par l'utilisateur tant donn sa spcificit de
  logiciel libre, qui peut le rendre complexe  manipuler et qui le rserve donc
   des dveloppeurs et des professionnels avertis possdant des connaissances
  informatiques approfondies.  Les utilisateurs sont donc invits  charger et
  tester l'adquation du logiciel  leurs besoins dans des conditions permettant
  d'assurer la scurit de leurs systmes et ou de leurs donnes et, plus
  gnralement,  l'utiliser et l'exploiter dans les mmes conditions de
  scurit.

  Le fait que vous puissiez accder  cet en-tte signifie que vous avez pris
  connaissance de la licence CeCILL, et que vous en avez accept les termes.
*/

/**
  * \file hdf_get_obj_att.c
  * \brief Rcupration des attributs d'un objet HDF
  *
  * Cette fonction retourne une structure dcrivant les attribut d'un objet HDF.
  * Voir hdfatt.h pour la dfinition de cette structure.
  *
  * Elle retourne 0 en cas de succs, 1 en cas d'erreur HDF et 2 pour toute
  * autre erreur
  * ----------------------------------------------------------------------------
  *
  * Spcification: int hdf_get_obj_att(int32 obj_id, int32 nbatt,
  *                                    struct hdfatt **att);
  *
  * ----------------------------------------------------------------------------
  *
  * \author SIX Bruno, CGTD ICARE
  * \author CRI USTL, Domaine Universitaire Scientifique
  * \version 1.0.0
  * \date 2006-06-09
  *
  * ----------------------------------------------------------------------------
  *
  * <b>Modifications, rvisions:</b>
  *
  * ----------------------------------------------------------------------------
  *
  */

#include <hdfatt.h>

int hdf_get_obj_att(
  /** Identificateur HDF de l'objet                    */ int32         obj_id,
  /** Nombre d'attributs de l'objet                    */ int32         nbatt,
  /** Pointeur sur la table des attributs  renseigner */ struct hdfatt **att  )
{
  int32 iatt;    /* Indice de l'attribut courant */
  int8 *attdat;  /* Zone mmoire pour la rcupration des valeurs de l'attribut courant */
  
  /* Allocation de la table d'attributs pour cet objet */
  if ((*att = (struct hdfatt *)malloc(nbatt*sizeof(struct hdfatt))) == NULL) goto ERR_STD;
  
  /* DEBUT DE BOUCLE SUR LES ATTRIBUTS A EXTRAIRE */
  for (iatt = 0; iatt < nbatt; iatt++) {

    /* Rcupration des informations sur l'attribut courant */
    if (SDattrinfo(obj_id, iatt, (*att + iatt)->name, &((*att + iatt)->typ),
		   &((*att + iatt)->dim)) < 0) goto ERR_HDF;

    /* Alloc. de la zone mmoire pour l'extraction des valeurs de l'attribut */
    if ((attdat = (int8 *)malloc((1+(*att + iatt)->dim)*DFKNTsize((*att + iatt)->typ)))
	== NULL) goto ERR_STD;
    attdat[(*att + iatt)->dim] = 0;

    /* Extraction des valeur de cet attribut */
    if (SDreadattr(obj_id, iatt, attdat) < 0) goto ERR_HDF;
    (*att + iatt)->data = (VOIDP)attdat;
  }

  /* Fin normale de la fonction */
  return 0;

 ERR_HDF:   /* Fin de la fonction sur une erreur HDF locale */
  HEprint(stderr, 0);
  return 1;

 ERR_STD:   /* Fin de la fonction sur une autre erreur locale */
  perror("ERROR");
  return 2;
}
